from .fragments import ASSET_FRAGMENT


GQL_APPEND_MANY_TO_DATASET = f'''
mutation(
    $projectID: ID!,
    $contentArray: [String!],
    $externalIDArray: [String!],
    $isHoneypotArray: [Boolean!],
    $statusArray: [Status!],
    $jsonMetadataArray: [String!]) {{
  data: appendManyToDataset(
    projectID: $projectID,
    contentArray: $contentArray,
    externalIDArray: $externalIDArray,
    isHoneypotArray: $isHoneypotArray,
    statusArray: $statusArray,
    jsonMetadataArray: $jsonMetadataArray
  ) {{
    {ASSET_FRAGMENT}
  }}
}}
'''


GQL_UPDATE_PROPERTIES_IN_ASSET = f'''
mutation(
    $assetID: ID!
    $consensusMark: Float
    $content: String
    $externalId: String
    $honeypotMark: Float
    $isHoneypot: Boolean
    $isUsedForConsensus: Boolean
    $jsonMetadata: String
    $priority: Int
    $shouldResetToBeLabeledBy: Boolean
    $status: Status
    $toBeLabeledBy: [String]
) {{
  data: updatePropertiesInAsset(
    where: {{id: $assetID}},
    data: {{
      consensusMark: $consensusMark
      content: $content
      externalId: $externalId
      honeypotMark: $honeypotMark
      isHoneypot: $isHoneypot
      isUsedForConsensus: $isUsedForConsensus
      jsonMetadata: $jsonMetadata
      priority: $priority
      shouldResetToBeLabeledBy: $shouldResetToBeLabeledBy
      status: $status
      toBeLabeledBy: $toBeLabeledBy
    }}
  ) {{
    {ASSET_FRAGMENT}
  }}
}}
'''

GQL_DELETE_MANY_FROM_DATASET = f'''
mutation($assetIDs: [ID!]) {{
  data: deleteManyFromDataset(assetIDs: $assetIDs) {{
    {ASSET_FRAGMENT}
  }}
}}
'''



import kili.mutations.asset
import kili.mutations.label
import kili.mutations.notification
import kili.mutations.organization
import kili.mutations.project
import kili.mutations.user
import kili.queries.asset
import kili.queries.label
import kili.queries.organization
import kili.queries.project
import kili.queries.project_user
import kili.queries.user
import kili.subscriptions.label
from kili.constants import NO_ACCESS_RIGHT

from kili.helpers import deprecate


class Playground(
        kili.mutations.asset.MutationsAsset,
        kili.mutations.label.MutationsLabel,
        kili.mutations.notification.MutationsNotification,
        kili.mutations.organization.MutationsOrganization,
        kili.mutations.project.MutationsProject,
        kili.mutations.user.MutationsUser,
        kili.queries.asset.QueriesAsset,
        kili.queries.label.QueriesLabel,
        kili.queries.organization.QueriesOrganization,
        kili.queries.project.QueriesProject,
        kili.queries.project_user.QueriesProjectUser,
        kili.queries.user.QueriesUser,
        kili.subscriptions.label.SubscriptionsLabel):

    def __init__(self, auth=None):
        """Create an instance of KiliPlayground."""
        self.auth = auth
        super().__init__(auth)


if __name__ == '__main__':
    """ Example of usage """
    from kili.authentication import KiliAuth
    from kili.playground import Playground
    kauth = KiliAuth()
    playground = Playground(kauth)
    assets = playground.assets(project_id="first-project")
    print(assets)



from json import dumps
from typing import List

from ...helpers import format_result
from .queries import (GQL_APPEND_TO_LABELS, GQL_CREATE_HONEYPOT,
                      GQL_CREATE_PREDICTIONS,
                      GQL_UPDATE_PROPERTIES_IN_LABEL)


class MutationsLabel:

    def __init__(self, auth):
        """
        Initializes the subclass

        Parameters
        ----------
        - auth : KiliAuth object
        """
        self.auth = auth

    def create_predictions(self, project_id: str, external_id_array: List[str], model_name_array: List[str], json_response_array: List[dict]):
        """
        Create predictions for some assets

        Parameters
        ----------
        - project_id : str
        - external_id_array : list of str
            The external identifiers of the assets for which we want to add predictions
        - model_name_array : list of str
            In case you want to precise from which model the label originated
        - json_response_array : list of dict
            The predictions are given here. An example can be found here, for a polygon
            ```
            {
                "JOB_0": {
                    "annotations": [
                        {
                            "categories": [{ "name": "OBJECT_B", "confidence": 100 }],
                            "boundingPoly": [
                                {
                                    "normalizedVertices": [
                                        { "x": 0.07, "y": 0.88 },
                                        { "x": 0.07, "y": 0.32 },
                                        { "x": 0.94, "y": 0.32 },
                                        { "x": 0.94, "y": 0.88 }
                                    ]
                                }
                            ]
                        }
                    ]
                }
            }
            ```

        Returns
        -------
        - a result object which indicates if the mutation was successful, or an error message else.
        """
        assert len(external_id_array) == len(
            json_response_array), "IDs list and predictions list should have the same length"
        assert len(external_id_array) == len(
            model_name_array), "IDs list and model names list should have the same length"
        variables = {
            'projectID': project_id,
            'externalIDArray': external_id_array,
            'modelNameArray': model_name_array,
            'jsonResponseArray': [dumps(elem) for elem in json_response_array]
        }
        result = self.auth.client.execute(GQL_CREATE_PREDICTIONS, variables)
        return format_result('data', result)


    def append_to_labels(self, author_id: str, json_response: dict, label_asset_id: str, label_type: str, seconds_to_label: int, skipped: bool = False):
        """
        Append a label to labels

        Parameters
        ----------
        - author_id : str
            Email of the author of the label
        - json_response : dict
            Label is given here
        - label_asset_id : str
            Identifier of the asset
        - label_type : str
            Can be one of {'AUTOSAVE', 'DEFAULT', 'PREDICTION', 'REVIEW'}
        - seconds_to_label : int
        - skipped : bool, optional (default = False)

        Returns
        -------
        - a result object which indicates if the mutation was successful, or an error message else.
        """
        variables = {
            'authorID': author_id,
            'jsonResponse': dumps(json_response),
            'labelAssetID': label_asset_id,
            'labelType': label_type,
            'secondsToLabel': seconds_to_label,
            'skipped': skipped
        }
        result = self.auth.client.execute(GQL_APPEND_TO_LABELS, variables)
        return format_result('data', result)


    def update_properties_in_label(self, label_id: str, seconds_to_label: int = None, model_name: str = None, json_response: dict = None):
        """
        Update properties of a label

        Parameters
        ----------
        - label_id : str
            Identifier of the label
        - seconds_to_label : int, optional (default = None)
        - model_name : str, optional (default = None)
        - json_response : dict, optional (default = None)
            The label is given here

        Returns
        -------
        - a result object which indicates if the mutation was successful, or an error message else.
        """
        formatted_json_response = None if json_response is None else dumps(
            json_response)
        variables = {
            'labelID': label_id,
            'secondsToLabel': seconds_to_label,
            'modelName': model_name,
            'jsonResponse': formatted_json_response
        }
        result = self.auth.client.execute(GQL_UPDATE_PROPERTIES_IN_LABEL, variables)
        return format_result('data', result)


    def create_honeypot(self, asset_id: str, json_response: dict):
        """
        Create honeypot for an asset. 
        
        Uses the json_response given to create a "REVIEW" label. This allows to compute a `honeypotMark`,
        which measures how similar are other labels compared to this one.

        Parameters
        ----------
        - asset_id : str
            Identifier of the asset
        - json_response : dict
            The honeypot label of the asset

        Returns
        -------
        - a result object which indicates if the mutation was successful, or an error message else.
        """
        variables = {
            'assetID': asset_id,
            'jsonResponse': dumps(json_response)
        }
        result = self.auth.client.execute(GQL_CREATE_HONEYPOT, variables)
        return format_result('data', result)


import functools
import warnings
import base64
import re
from json import dumps, loads
from types import *


class GraphQLError(Exception):
    def __init__(self, mutation, error):
        super().__init__(f'Mutation "{mutation}" failed with error: "{error}"')


def format_result(name, result):
    if 'errors' in result:
        raise GraphQLError(name, result['errors'])
    return format_json(result['data'][name])


def content_escape(content):
    return content.replace('\\', '\\\\').replace('\n', '\\n').replace('"', '\\"')


def get_data_type(path):
    if re.match(r'(jpg|jpeg)$', path.lower()):
        return 'image/png'


def encode_image(path):
    data_type = get_data_type(path)
    with open(path, 'rb') as image_file:
        return f'data:{data_type};base64,' + \
            base64.b64encode(image_file.read()).decode('ascii')


def is_url(path):
    return re.match(r'^(http://|https://)', path.lower())


def format_json(result):
    if result is None:
        return result
    if isinstance(result, list):
        return [format_json(elem) for elem in result]
    if isinstance(result, dict):
        for key, value in result.items():
            if key in ['jsonInterface', 'jsonMetadata', 'jsonResponse']:
                if value == '' or value is None:
                    result[key] = dict()
                else:
                    try:
                        result[key] = loads(value)
                    except:
                        raise ValueError(
                            'Json Metadata / json response / json interface should be valid jsons')
            else:
                result[key] = format_json(value)
        return result
    return result


def fragment_builder(fields, type_of_fields):
    fragment = ''
    subfields = [field.split('.', 1) for field in fields if '.' in field]
    if subfields:
        for subquery in set([subfield[0] for subfield in subfields]):
            type_of_fields_subquery = getattr(type_of_fields, subquery)
            try:
                if issubclass(type_of_fields_subquery, object):
                    fields_subquery = [subfield[1]
                                       for subfield in subfields if subfield[0] == subquery]
                    fragment += f' {subquery}{{{fragment_builder(fields_subquery,type_of_fields_subquery)}}}'
            except ValueError:
                print(f'{subquery} must be a valid subquery field')
        fields = [field for field in fields if '.' not in field]
    for field in fields:
        try:
            getattr(type_of_fields, field)
        except ValueError:
            print(f'{field} must be an instance of {type_of_fields}')
        if isinstance(field, str):
            fragment += f' {field}'
        else:
            raise Exception('Please provide the fields to query as strings')
    return fragment


def deprecate(msg, type=DeprecationWarning):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(msg, type, stacklevel=2)
            return func(*args, **kwargs)
        return wrapper
    return decorator



import os
import warnings

import requests

from . import __version__
from .graphql_client import GraphQLClient
from .mutations.user import signin

warnings.filterwarnings("default", module='kili', category=DeprecationWarning)

MAX_RETRIES = 20


def get_version_without_patch(version):
    return '.'.join(version.split('.')[:-1])


class KiliAuth(object):
    """
    from kili.authentication import KiliAuth
    from kili.playground import Playground
    kauth = KiliAuth(email=email, password=password)
    playground = Playground(kauth)
    assets = playground.assets(project_id=project_id)
    """

    def __init__(self,
                 email=os.getenv('KILI_USER_EMAIL'),
                 password=os.getenv('KILI_USER_PASSWORD'),
                 api_endpoint='https://cloud.kili-technology.com/api/label/graphql',
                 api_key=None,
                 verify=True):
        self.session = requests.Session()

        self.verify = verify

        self.check_versions_match(api_endpoint)

        adapter = requests.adapters.HTTPAdapter(max_retries=MAX_RETRIES)
        self.session.mount('https://', adapter)
        self.session.mount('http://', adapter)
        self.client = GraphQLClient(api_endpoint, self.session, verify=self.verify)
        if api_key is None:
            message = 'Authentication with a password is deprecated from 03/07/2020\n' + \
                      'It will be removed on 03/08/2020. Please generate an API key\n' + \
                      ' on the App in `My Account` and use `KiliAuth(api_key=...)` to connect'
            warnings.warn(message, DeprecationWarning)
            auth_payload = signin(self.client, email, password)
            api_token = auth_payload['token']
            self.client.inject_token('Bearer: ' + api_token)
            self.user_id = auth_payload['user']['id']
        else:
            self.client.inject_token('X-API-Key: ' + api_key)

    def __del__(self):
        self.session.close()

    def check_versions_match(self, api_endpoint):
        url = api_endpoint.replace('/graphql', '/version')
        response = requests.get(url, verify=self.verify).json()
        version = response['version']
        if get_version_without_patch(version) != get_version_without_patch(__version__):
            message = 'Kili Playground version should match with Kili API version.\n' + \
                      f'Please install version: "pip install kili=={version}"'
            warnings.warn(message, UserWarning)



from ...helpers import format_result
from .queries import (GQL_CREATE_NOTIFICATION,
                      GQL_UPDATE_PROPERTIES_IN_NOTIFICATION)


class MutationsNotification:

    def __init__(self, auth):
        """
        Initializes the subclass

        Parameters
        ----------
        - auth : KiliAuth object
        """
        self.auth = auth

    def create_notification(self, message: str, status: str, url: str, user_id: str):
        """
        Create a notification

        This method is currently only active for Kili administrators.

        Parameters
        ----------
        - message : str
        - status : str
        - url : str
        - user_id : str

        Returns
        -------
        - a result object which indicates if the mutation was successful, or an error message else.
        """
        variables = {
            'data': {
                'message': message,
                'status': status,
                'url': url,
                'userID': user_id
            }
        }
        result = self.auth.client.execute(GQL_CREATE_NOTIFICATION, variables)
        return format_result('data', result)

    def update_properties_in_notification(self, notification_id: str, status: str, url: str):
        """
        Modify a notification

        This method is currently only active for Kili administrators.

        Parameters
        ----------
        - notification_id : str
        - status : str
        - url : str

        Returns
        -------
        - a result object which indicates if the mutation was successful, or an error message else.
        """
        variables = {
            'id': notification_id,
            'status': status,
            'url': url
        }
        result = self.auth.client.execute(
            GQL_UPDATE_PROPERTIES_IN_NOTIFICATION, variables)
        return format_result('data', result)
